#ifndef _DIRECTED_GRAPH_h
#define _DIRECTED_GRAPH_h

#include <initializer_list>
#include <stdexcept>
#include <string>
#include <unordered_map>
#include <unordered_set>
#include <utility> // 'std::pair'

class Node {
public:
	const std::string name;
	Node(const std::string&);
	bool operator==(const Node&) const; // Distinction is based on 'name'...
};

class _Node_key {
public:
	std::size_t operator()(const Node&) const noexcept; // Hash function based on 'name'...
};

class Directed_graph {
private: // Edges: A->B, A->D, B->C, B->E, D->E, E->C and possible C->D cycle
	std::unordered_multimap<const Node, const Node, _Node_key> _representation; // (A->(B,C),(B->(C,E)),(D->(E)),(E->(C)) and possible (C->(D)) cycle
	bool _cyclic_(const Node&, std::pair<std::list<std::pair<const Node, const Node> >::const_iterator, std::list<std::pair<const Node, const Node> >::const_iterator>&, const std::unordered_set<Node, _Node_key>&, int&) const;
public:
	Directed_graph(const std::initializer_list<std::unordered_multimap<const Node, const Node, _Node_key>::value_type>&);
	bool acyclic_() const;
	inline std::size_t size() const { return _representation.size(); }
};

#endif
