#include <iostream>
#include <ranges> // Unused...

#include "Directed_graph.h"

Node::Node(const std::string& name) : name(name) {}
bool Node::operator==(const Node& node) const { return name == node.name; }
std::size_t _Node_key::operator()(const Node& node) const noexcept {
	return std::hash<std::string>{}(node.name); // Empty curly braces creates an object to instrument the call of '()' operator...
}

Directed_graph::Directed_graph(const std::initializer_list<std::unordered_multimap<const Node, const Node, _Node_key>::value_type>& edges) {
	for (auto& [edge_origin, edge_destination] : edges) { // C++17 structural bindings...
		auto range = _representation.equal_range(edge_origin); // All destinations from 'edge_origin'...
		auto addition = true;
		for (auto& i = range.first; addition && i != range.second; i++)
			if (edge_destination == (*i).second)
				addition = false; // Avoid 2 or several times, for instance, A->B
		if (addition)
			_representation.insert(std::make_pair(edge_origin, edge_destination)); // A->B, A->D, Etc.
	}
}
bool Directed_graph::_cyclic_(const Node& from, std::pair<std::list<std::pair<const Node, const Node> >::const_iterator, std::list<std::pair<const Node, const Node> >::const_iterator >& outgoing_edges, const std::unordered_set<Node, _Node_key>& edge_origins, int& cost) const {
	for (auto& i = outgoing_edges.first; i != outgoing_edges.second; i++) {
		// from = A *** outgoing_edges = {(A->B),(A->D}
		if (from == (*i).second)
			return true; // Loop like A->A
		// if (edge_origins.contains((*i).second)) { // C++20 only...
		if (edge_origins.find((*i).second) == edge_origins.end()) { // If B in 'edge_origins' then don't explore from A because A->B...
			auto outgoing_edges = _representation.equal_range((*i).second);
			cost++;
			if (_cyclic_(from, outgoing_edges, edge_origins, cost))
				return true; // If at least one cycle found then graph isn't acyclic...
		}
	}
	return false; // No cycle found yet...
}
bool Directed_graph::acyclic_() const {
	// If edges have the same origin then get origin once:
	std::unordered_set<Node, _Node_key> edge_origins;
	for (auto& [edge_origin, edge_destination] : _representation)
		edge_origins.insert(edge_origin);
	// std::list<Node> edge_origins;
	// std::ranges::copy(std::views::keys(_representation), std::back_inserter(edge_origins)); // C++20
	// Alternative with pipe syntax from C++20:
	// std::ranges::copy(_representation | std::views::keys, std::back_inserter(edge_origins));
	// edge_origins.unique(); // Non-robust because only *CONSECUTIVE* duplicates are removed...
	int cost = 0;
	for (auto& edge_origin : edge_origins) { // Explore graph from nodes as edge origins...
		auto edge_destinations = _representation.equal_range(edge_origin); // e.g., values of A key -> (B,C)
		if (_cyclic_(edge_origin, edge_destinations, edge_origins, cost))
			return false; // If at least one cycle found then graph isn't acyclic...
	}
	return cost != 0;
}

int main(int argc, char** argv) {
	const Node a("A"), b("B"), c("C"), d("D"), e("E");
	Directed_graph directed_graph({ /*std::make_pair(a, a),*/std::make_pair(a, b),std::make_pair(a, b),std::make_pair(a, d),
		std::make_pair(b, c),std::make_pair(b,e),
		std::make_pair(d, e),std::make_pair(e, c),/*std::make_pair(c, d)*/ });
	// std::cout << "'directed_graph.size()': " << directed_graph.size() << std::endl;
	std::cout << "Is 'directed_graph' acyclic? " << (directed_graph.acyclic_() ? "YES" : "NO") << std::endl;

	return 0;
}

